# -*- coding: utf-8 -*-
"""
Scrapbag geo file.
"""
import structlog
from pycountry import countries
from geopy.geocoders import Nominatim

logger = structlog.getLogger(__name__)


def get_country(similar=False, **kwargs):
    """
    Get a country for pycountry
    """
    result_country = None
    try:
        if similar:
            for country in countries:
                if kwargs.get('name', '') in country.name:
                    result_country = country
                    break
        else:
            result_country = countries.get(**kwargs)
    except Exception as ex:
        msg = ('Country not found in pycountry with params introduced'
               ' - {}'.format(ex))
        logger.error(msg, params=kwargs)

    return result_country


def get_location(address=""):
    """
    Retrieve location coordinates from an address introduced.
    """
    coordinates = None
    try:
        geolocator = Nominatim()
        location = geolocator.geocode(address)
        coordinates = (location.latitude, location.longitude)
    except Exception as ex:
        logger.error('Fail get location - {}'.format(ex))
    return coordinates


def get_address(coords=None, **kwargs):
    """
    Retrieve addres from a location in coords format introduced.
    """
    address = None
    try:
        if (not coords) and \
            ('latitude' in kwargs and 'longitude' in kwargs) or \
                ('location' in kwargs):

            coords = kwargs.get(
                'location', (kwargs.get('latitude'), kwargs.get('longitude')))

        # transform coords
        if isinstance(coords, (list, tuple)) and len(coords) == 2:
            coords = "{}, {}".join(map(str, coords))

        geolocator = Nominatim()
        location = geolocator.reverse(coords)
        address = location.address
    except Exception as ex:
        logger.error('Fail get reverse address - {}'.format(ex))
    return address



# -*- coding: utf-8 -*-
"""
Scrapbag censal file.
"""
import os
from .csvs import csv_tolist
from .strings import normalizer

import structlog

logger = structlog.getLogger(__name__)

DEFAULT_DATA_PATH = os.path.join(
    "/".join(os.path.abspath(__file__).split('/')[:-1]),
    'data-aux/18codmun.csv')


class IneCensal():

    index = {}

    def __init__(self, **kwargs):

        # load Ine file data
        lista = self.load_data(**kwargs)

        # generate the index
        self.index = self.make_index(lista)

    def load_data(self, path=DEFAULT_DATA_PATH, **kwargs):

        return csv_tolist(path, delimiter=';')

    def make_index(self, lista):

        ine_censo_index = {}

        # first row is headers
        for censo_data in lista[1:]:
            try:
                c_data = None

                # In ine file data, there are some rows with this pattern,
                # 'florida, La' and we want to change to 'La florida'

                if ',' in censo_data[4]:
                    c_data = " ".join(censo_data[4].split(', ')[::-1])
                else:
                    c_data = censo_data[4]

                ine_censo_index[normalizer(c_data)] = "".join(censo_data[1:3])

            except Exception as ex:
                logger.error('Fail to make index with parsed row data {} - {}'.format(censo_data, ex))

        return ine_censo_index

    def get_codigo_censal(self, localidad_name, without_codigo_censal=set()):

        # assert input name normalizer
        localidad_name = normalizer(localidad_name)

        codigo_censal = self.index.get(localidad_name, None)

        if not codigo_censal:
            result_key = None
            search_data = list(self.index.keys())

            if localidad_name not in without_codigo_censal:

                for name_word in localidad_name.split('-'):
                    logger.debug('*Searching {}'.format(localidad_name))

                    if len(search_data) == 1:

                        # same word lenght
                        if len(search_data[0].split('-')) == len(localidad_name.split('-')):
                            result_key = search_data[0]
                            logger.debug('*** Found codigo censal: {}'.format(result_key))

                        break

                    if not search_data:
                        break

                    for search_name in iter(search_data):
                        if name_word not in search_name:
                            search_data.remove(search_name)

            if result_key is None:
                without_codigo_censal.add(localidad_name)
                logger.error('* -- Fail Searching codigo_censal {}'.format(localidad_name))

        return codigo_censal



# -*- coding: utf-8 -*-
"""
Scrapbag dates file.
"""
import re
import locale
import calendar

from datetime import timedelta, datetime
from dateutil.relativedelta import relativedelta


import pytz
import structlog


logger = structlog.getLogger(__name__)

LOCALES = {
    "es": locale.normalize("es_ES.UTF-8"),
    "en": locale.normalize("en_US.UTF-8"),
    "fr": locale.normalize("fr_FR.UTF-8"),
    "de": locale.normalize("de_DE.UTF-8"),
    "it": locale.normalize("it_IT.UTF-8"),
}

DEFAULT_DATE_LANG = 'en'


def get_dates_in_period(start=None, top=None, step=1, step_dict={}):
    """Return a list of dates from the `start` to `top`."""

    delta = relativedelta(**step_dict) if step_dict else timedelta(days=step)

    start = start or datetime.today()
    top = top or start + delta
    dates = []
    current = start
    while current <= top:
        dates.append(current)
        current += delta
    return dates


def ordinal_suffix(day):
    """Return ordinal english suffix to number of a day."""
    condition = 4 <= day <= 20 or 24 <= day <= 30
    return 'th' if condition else ["st", "nd", "rd"][day % 10 - 1]


def localize_date(date, city):
    """ Localize date into city

    Date: datetime
    City: timezone city definitio. Example: 'Asia/Qatar', 'America/New York'..
    """
    local = pytz.timezone(city)
    local_dt = local.localize(date, is_dst=None)
    return local_dt


def get_year_from_date_str(date_str):
    """
    Retrieve only the year from a text string.
    """
    return re.findall(r'\d{4}', date_str) if date_str else None


def get_month_from_date_str(date_str, lang=DEFAULT_DATE_LANG):
    """Find the month name for the given locale, in the given string.

    Returns a tuple ``(number_of_month, abbr_name)``.
    """
    date_str = date_str.lower()
    with calendar.different_locale(LOCALES[lang]):
        month_abbrs = list(calendar.month_abbr)
        for seq, abbr in enumerate(month_abbrs):
            if abbr and abbr.lower() in date_str:
                return seq, abbr
    return ()


def replace_month_abbr_with_num(date_str, lang=DEFAULT_DATE_LANG):
    """Replace month strings occurrences with month number."""
    num, abbr = get_month_from_date_str(date_str, lang)
    return re.sub(abbr, str(num), date_str, flags=re.IGNORECASE)


def translate_month_abbr(
        date_str,
        source_lang=DEFAULT_DATE_LANG,
        target_lang=DEFAULT_DATE_LANG):
    """Translate the month abbreviation from one locale to another."""
    month_num, month_abbr = get_month_from_date_str(date_str, source_lang)
    with calendar.different_locale(LOCALES[target_lang]):
        translated_abbr = calendar.month_abbr[month_num]
        return re.sub(
            month_abbr, translated_abbr, date_str, flags=re.IGNORECASE)


def merge_datetime(date, time='', date_format='%d/%m/%Y', time_format='%H:%M'):
    """Create ``datetime`` object from date and time strings."""
    day = datetime.strptime(date, date_format)
    if time:
        time = datetime.strptime(time, time_format)
        time = datetime.time(time)
        day = datetime.date(day)
        day = datetime.combine(day, time)
    return day



# -*- coding: utf-8 -*-
"""
Scrapbag csv file.
"""
import io
import csv
import codecs
import itertools
import collections
import xlrd
import structlog


from .strings import normalizer
from .collections import (exclude_empty_values, remove_list_duplicates,
                          force_list)
logger = structlog.getLogger(__name__)

# Output formats
ARRAY_RAW_FORMAT = 0
ARRAY_CLEAN_FORMAT = 1
DICT_FORMAT = 2


def get_csv_col_headers(rows, row_headers_count_value=0):
    """
    Retrieve csv column headers
    """
    count = 0

    if rows:
        for row in rows:
            if exclude_empty_values(row[:row_headers_count_value]):
                break
            count += 1

    if len(rows) == count:
        count = 1  # by default

    return [r[row_headers_count_value:] for r in rows[:count]]


def row_headers_count(rows):
    """
    Count the row headers in rows
    """
    count = 0

    if rows:
        for data in rows[0]:
            if not data:
                count += 1
            else:
                break
    return count


def populate_csv_headers(rows,
                         partial_headers,
                         column_headers_count=1):
    """
    Populate csv rows headers when are empty, extending the superior or
    upper headers.
    """

    result = [''] * (len(rows) - column_headers_count)

    for i_index in range(0, len(partial_headers)):
        for k_index in range(0, len(partial_headers[i_index])):

            # missing field find for a value in upper rows
            if not partial_headers[i_index][k_index] and i_index - 1 >= 0:

                # TODO: It's necesary a for or only taking the
                # inmediate latest row works well??
                for t_index in range(i_index - 1, -1, -1):
                    # TODO: could suposse that allways a value exists
                    partial_value = partial_headers[t_index][k_index]
                    if partial_value:
                        partial_headers[i_index][k_index] = partial_value
                        break

        result[i_index] = " ".join(map(str, partial_headers[i_index]))

    return result


def get_row_headers(rows, row_headers_count_value=0, column_headers_count=1):
    """
    Return row headers.
    Assume that by default it has one column header.
    Assume that there is only one father row header.
    """
    # TODO: REFACTOR ALGORITHM NEEDED
    partial_headers = []

    if row_headers_count_value:

        # Take partial data
        for k_index in range(0, len(rows) - column_headers_count):
            header = rows[k_index + column_headers_count][
                :row_headers_count_value]
            partial_headers.append(remove_list_duplicates(force_list(header)))

        # Populate headers
        populated_headers = populate_csv_headers(
            rows,
            partial_headers,
            column_headers_count)

        return populated_headers


def retrieve_csv_data(rows, row_header=0, column_header=0, limit_column=0):
    """
    Take the data from the rows.
    """
    return [row[row_header:limit_column] for row in rows[column_header:]]


def csv_tolist(path_to_file, **kwargs):
    """
    Parse the csv file to a list of rows.
    """

    result = []

    encoding = kwargs.get('encoding', 'utf-8')
    delimiter = kwargs.get('delimiter', ',')
    dialect = kwargs.get('dialect', csv.excel)

    _, _ext = path_to_file.split('.', 1)

    try:

        file = codecs.open(path_to_file, 'r', encoding)
        items_file = io.TextIOWrapper(file, encoding=encoding)
        result = list(
            csv.reader(items_file, delimiter=delimiter, dialect=dialect))

        items_file.close()
        file.close()

    except Exception as ex:
        result = []
        logger.error('Fail parsing csv to list of rows - {}'.format(ex))

    return result


def excel_todictlist(path_to_file, **kwargs):
    """
    Parse excel file to a dict list of sheets, rows.
    """
    result = collections.OrderedDict()
    encoding = kwargs.get('encoding', 'utf-8')
    formatting_info = '.xlsx' not in path_to_file
    count = 0

    with xlrd.open_workbook(
        path_to_file,
        encoding_override=encoding, formatting_info=formatting_info) \
            as _excelfile:

        for sheet_name_raw in _excelfile.sheet_names():

            # if empty sheet name put sheet# as name
            sheet_name = sheet_name_raw or "sheet{}".format(count)
            result[sheet_name] = []

            xl_sheet = _excelfile.sheet_by_name(sheet_name_raw)

            for row_idx in range(0, xl_sheet.nrows):
                col_data = []
                for col_idx in range(0, xl_sheet.ncols):

                    # Get cell object by row, col
                    cell_obj = xl_sheet.cell(row_idx, col_idx)
                    merged_info = is_merged(xl_sheet, row_idx, col_idx)

                    # Search for value in merged_info
                    if not cell_obj.value and merged_info:
                        cell_obj = search_mergedcell_value(
                            xl_sheet, merged_info[1])
                        col_data.append(cell_obj.value if cell_obj else '')
                    else:
                        col_data.append(cell_obj.value)

                result[sheet_name].append(col_data)

            count += 1  # increase sheet counter

    return result


def search_mergedcell_value(xl_sheet, merged_range):
    """
    Search for a value in merged_range cells.
    """
    for search_row_idx in range(merged_range[0], merged_range[1]):
        for search_col_idx in range(merged_range[2], merged_range[3]):
            if xl_sheet.cell(search_row_idx, search_col_idx).value:
                return xl_sheet.cell(search_row_idx, search_col_idx)
    return False


def is_merged(sheet, row, column):
    """
    Check if a row, column cell is a merged cell
    """
    for cell_range in sheet.merged_cells:
        row_low, row_high, column_low, column_high = cell_range
        if (row in range(row_low, row_high)) and \
                (column in range(column_low, column_high)):

            # TODO: IS NECESARY THIS IF?
            if ((column_high - column_low) < sheet.ncols - 1) and \
                    ((row_high - row_low) < sheet.nrows - 1):
                return (True, cell_range)

    return False


def populate_headers(headers):
    """
    Concatenate headers with subheaders
    """
    result = [''] * len(headers[0])
    values = [''] * len(headers)
    for k_index in range(0, len(headers)):
        for i_index in range(0, len(headers[k_index])):
            if headers[k_index][i_index]:
                values[k_index] = normalizer(
                    str(headers[k_index][i_index]))  # pass to str

            if len(exclude_empty_values(result)) > i_index:
                result[i_index] += "-{}".format(values[k_index])
            else:
                result[i_index] += str(values[k_index])

    return result


def row_csv_limiter(rows, limits=None):
    """
    Limit row passing a value or detect limits making the best effort.
    """

    limits = [None, None] if limits is None else limits

    if len(exclude_empty_values(limits)) == 2:
        upper_limit = limits[0]
        lower_limit = limits[1]
    elif len(exclude_empty_values(limits)) == 1:
        upper_limit = limits[0]
        lower_limit = row_iter_limiter(rows, 1, -1, 1)
    else:
        upper_limit = row_iter_limiter(rows, 0, 1, 0)
        lower_limit = row_iter_limiter(rows, 1, -1, 1)

    return rows[upper_limit: lower_limit]


def row_iter_limiter(rows, begin_row, way, c_value):
    """
    Alghoritm to detect row limits when row have more that one column.
    Depending the init params find from the begin or behind.
    NOT SURE THAT IT WORKS WELL..
    """
    limit = None

    for index in range(begin_row, len(rows)):
        if not len(exclude_empty_values(rows[way * index])) == 1:
            limit = way * index + c_value if way * index + \
                c_value not in [way * len(rows), 0] else None
            break

    return limit


def csv_row_cleaner(rows):
    """
    Clean row checking:
     - Not empty row.
     - >=1 element different in a row.
     - row allready in cleaned row result.


    """
    result = []

    for row in rows:

        # check not empty row
        check_empty = len(exclude_empty_values(row)) > 1

        # check more or eq than 1 unique element in row
        check_set = len(set(exclude_empty_values(row))) > 1
        # check row not into result cleaned rows.
        check_last_allready = (result and result[-1] == row)

        if check_empty and check_set and not check_last_allready:
            result.append(row)
    return result


def csv_column_cleaner(rows):
    """
    clean csv columns parsed omitting empty/dirty rows.
    """

    # check columns if there was empty columns
    result = [[] for x in range(0, len(rows))]
    for i_index in range(0, len(rows[0])):

        partial_values = []

        for x_row in rows:
            partial_values.append(
                x_row[i_index] if len(x_row) > i_index else '')

        colum_rows = exclude_empty_values(partial_values)

        if len(colum_rows) > len(rows) / 5:  # adjust this value
            for index in range(0, len(rows)):
                result[index].append(
                    rows[index][i_index] if len(rows[index]) > i_index else '')
    return result


def csv_column_header_cleaner(rows):
    """
    Clean column headers rows excluding empty values.
    """
    return exclude_empty_values(rows)


def csv_dict_format(csv_data, c_headers=None, r_headers=None):
    """
    Format csv rows parsed to Dict.
    """
    # format dict if has row_headers
    if r_headers:
        result = {}
        for k_index in range(0, len(csv_data)):
            if r_headers[k_index]:
                result[r_headers[k_index]] = collections.OrderedDict(
                    zip(c_headers, csv_data[k_index]))

    # format list if hasn't row_headers -- square csv
    else:
        result = []
        for k_index in range(0, len(csv_data)):
            result.append(
                collections.OrderedDict(zip(c_headers, csv_data[k_index])))
        result = [result]

    return result


def csv_array_clean_format(csv_data, c_headers=None, r_headers=None):
    """
    Format csv rows parsed to Array clean format.
    """

    result = []
    real_num_header = len(force_list(r_headers[0])) if r_headers else 0
    result.append([""] * real_num_header + c_headers)

    for k_index in range(0, len(csv_data)):

        if r_headers:
            result.append(
                list(
                    itertools.chain(
                        [r_headers[k_index]],
                        csv_data[k_index])))

        else:
            result.append(csv_data[k_index])

    return result


def csv_format(csv_data, c_headers=None, r_headers=None, rows=None, **kwargs):
    """
    Format csv rows parsed to Dict or Array
    """
    result = None
    c_headers = [] if c_headers is None else c_headers
    r_headers = [] if r_headers is None else r_headers
    rows = [] if rows is None else rows

    result_format = kwargs.get('result_format', ARRAY_RAW_FORMAT)

    # DICT FORMAT
    if result_format == DICT_FORMAT:
        result = csv_dict_format(csv_data, c_headers, r_headers)

    # ARRAY_RAW_FORMAT
    elif result_format == ARRAY_RAW_FORMAT:
        result = rows

    # ARRAY_CLEAN_FORMAT
    elif result_format == ARRAY_CLEAN_FORMAT:
        result = csv_array_clean_format(csv_data, c_headers, r_headers)

    else:
        result = None

    # DEFAULT
    if result and result_format < DICT_FORMAT:
        result = [result]

    return result


def csv_to_dict(csv_filepath, **kwargs):
    """
    Turn csv into dict.
    Args:
        :csv_filepath: path to csv file to turn into dict.
        :limits: path to csv file to turn into dict
    """
    callbacks = {'to_list': csv_tolist,
                 'row_csv_limiter': row_csv_limiter,
                 'csv_row_cleaner': csv_row_cleaner,
                 'row_headers_count': row_headers_count,
                 'get_col_header': get_csv_col_headers,
                 'get_row_headers': get_row_headers,
                 'populate_headers': populate_headers,
                 'csv_column_header_cleaner': csv_column_header_cleaner,
                 'csv_column_cleaner': csv_column_cleaner,
                 'retrieve_csv_data': retrieve_csv_data}

    callbacks.update(kwargs.get('alt_callbacks', {}))
    rows = kwargs.get('rows', [])

    if not rows:
        # csv_tolist of rows
        rows = callbacks.get('to_list')(csv_filepath, **kwargs)

        if not rows:
            msg = 'Empty rows obtained from {}'.format(csv_filepath)
            logger.warning(msg)
            raise ValueError(msg)

    # apply limits
    rows = callbacks.get('row_csv_limiter')(
        rows, kwargs.get('limits', [None, None]))

    # apply row cleaner
    rows = callbacks.get('csv_row_cleaner')(rows)

    # apply column cleaner
    rows = callbacks.get('csv_column_cleaner')(rows)

    # count raw headers
    num_row_headers = callbacks.get('row_headers_count')(rows)

    # take colum_headers
    c_headers_raw = callbacks.get('get_col_header')(rows, num_row_headers)

    # get row_headers
    r_headers = callbacks.get('get_row_headers')(
        rows, num_row_headers, len(c_headers_raw))

    # format colum_headers
    c_headers_dirty = callbacks.get('populate_headers')(
        c_headers_raw) if len(c_headers_raw) > 1 else c_headers_raw[0]

    # Clean csv column headers of empty values.
    c_headers = callbacks.get('csv_column_header_cleaner')(c_headers_dirty)

    # take data
    csv_data = callbacks.get('retrieve_csv_data')(
        rows,
        column_header=len(c_headers_raw),
        row_header=num_row_headers,
        limit_column=len(c_headers) - len(c_headers_dirty) or None)

    # Check column headers validation
    if csv_data:
        assert len(c_headers) == len(csv_data[0])

    # Check row headers validation
    if r_headers:
        assert len(r_headers) == len(csv_data)

    # Transform rows into dict zipping the headers.
    kwargs.pop('rows', None)
    result = csv_format(csv_data, c_headers, r_headers, rows, **kwargs)

    return result


def excel_to_dict(excel_filepath, encapsulate_filepath=False, **kwargs):
    """
    Turn excel into dict.
    Args:
        :excel_filepath: path to excel file to turn into dict.
        :limits: path to csv file to turn into dict
    """
    result = {}
    try:
        callbacks = {'to_dictlist': excel_todictlist}  # Default callback
        callbacks.update(kwargs.get('alt_callbacks', {}))

        # Retrieve excel data as dict of sheets lists
        excel_data = callbacks.get('to_dictlist')(excel_filepath, **kwargs)
        for sheet in excel_data.keys():
            try:
                kwargs['rows'] = excel_data.get(sheet, [])
                result[sheet] = csv_to_dict(excel_filepath, **kwargs)
            except Exception as ex:
                logger.error('Fail to parse sheet {} - {}'.format(sheet, ex))
                result[sheet] = []
                continue

        if encapsulate_filepath:
            result = {excel_filepath: result}

    except Exception as ex:
        msg = 'Fail transform excel to dict - {}'.format(ex)
        logger.error(msg, excel_filepath=excel_filepath)

    return result



# -*- coding: utf-8 -*-
"""
Test Scrapbag strings file
"""
import unittest
import collections

from scrapbag.collections import format_dict
from scrapbag.strings import (
    clean_text,
    clean_markdown,
    get_only_words,
    select_regexp_char,
    exclude_chars,
    normalizer,
    normalize_dict
    )


class UtilsStringsTestCase(unittest.TestCase):
    """
    Test Scrapbag test case
    """

    def test_clean_text(self):
        """
        Test clean_text
        """
        self.assertEqual(clean_text('    The   Cat   '), 'The Cat')
        self.assertEqual(clean_text(''), None)

    def test_clean_markdown(self):
        """
        Test clean_markdown
        """
        markdown_text = '####**  The   Cat  ** '
        self.assertEqual(clean_text(markdown_text), 'The Cat')
        self.assertEqual(clean_markdown([markdown_text]), [markdown_text])

    def test_get_only_words(self):
        """
        Test get_only_words
        """
        self.assertEqual(
            get_only_words('####**  The   Cat  ** '), ['The', 'Cat'])

    def test_select_regexp_char(self):
        """
        Test select_regexp_char
        """
        result = select_regexp_char('*')
        result2 = select_regexp_char('2')
        result3 = select_regexp_char([])
        result4 = select_regexp_char('a')

        self.assertEqual(result, r'\*')
        self.assertEqual(result2, '2')
        self.assertEqual(result3, '')
        self.assertEqual(result4, 'a')

    def test_exclude_chars(self):
        """
        Test exclude_chars
        """
        text = '-h**a*2!!loa12_'

        result1 = exclude_chars(text)
        result2 = exclude_chars(text, ['-', '*', '!', 'o', '1', '2', '_'])

        self.assertEqual(result1, text)
        self.assertEqual(result2, 'hala')

    def test_normalizer(self):
        """
        Test normalizer
        """
        text = 'Im @ te_[st;*ed] / `..+.  - "'
        exclusions = [' ', ',', '.', ';',
                      '@', "\\", "'", '"', "`", "[", "]", '\n']
        test_text = normalizer(text, exclusion=[])
        exclude_default = normalizer(text)

        self.assertIn('+', exclude_default)
        self.assertIn('/', exclude_default)
        self.assertIn('_', exclude_default)

        self.assertFalse(any(exc in test_text for exc in exclusions))
        self.assertNotEqual('-', test_text[-1])
        self.assertNotEqual('-', test_text[0])
        self.assertFalse(any([x.isupper() for x in test_text]))
        self.assertTrue(any([x.isupper() for x in normalizer(text, lower=False)]))

        first_step_text = normalizer(
            text, exclusion=['@'], lower=False, separate_char='')

        test_final_text_clean = normalizer(
            first_step_text, lower=False, exclusion=[], separate_char=' ')

        self.assertEqual(test_final_text_clean, 'Im tested')

    def test_normalizer_dict(self):
        """
        Test normalizer_dict
        """
        test_dict = collections.OrderedDict(
            {'a!a': 1, '*asd231__dab': 2, '-2c**Ç21)=': 3})
        test_dict2 = collections.OrderedDict(
            {'a!a': {
                '-a**q231¨a': 1,
                'ab': {
                    ')(ab  as(a': 5}},
             'b': 2, 'c': 3, 'd(pepe)': {'da': 7}})

        result1 = normalize_dict(test_dict, exclusion=['_'])
        result2 = normalize_dict(test_dict2)

        self.assertEqual(result1, {'a-a': 1, 'asd231__dab': 2, '2c-c21': 3})
        self.assertEqual(result2, {
            'b': 2, 'a-a': {
                'ab': {'ab-as-a': 5},
                'a**q231-a': 1},
            'c': 3, 'd-pepe': {'da': 7}})

    def test_format_dict(self):
        """
        Test format_dict
        """
        test = {'a': 1, 'b': 2, 'c': 4}
        format_list = ['a', 'b', 'c']

        result = format_dict(test, format_list)
        format_list.reverse()
        result2 = format_dict(test, format_list, '*')

        self.assertEqual(result, '1,2,4')
        self.assertEqual(result2, '4*2*1')



# -*- coding: utf-8 -*-
"""
Test Scrapbag geo file
"""
import unittest
from unittest.mock import MagicMock
import mock

from scrapbag.geo import (
    get_country,
    get_location,
    get_address)


class GeoTestCase(unittest.TestCase):
    """
    Scrapbag geo Test Case
    """

    def test_get_country(self):
        """
        Test get_country
        """
        spain = get_country(name='Spain')
        spain2 = get_country(alpha_2='ES')
        spain3 = get_country(alpha_3='ESP')
        czech = get_country(name='Czech', similar=True)

        bad = get_country(alpha_2='ESP')
        bad_similar = get_country(name='xxxlalaltest', similar=True)

        self.assertEqual(spain.alpha_2, 'ES')
        self.assertEqual(spain.alpha_3, 'ESP')
        self.assertEqual(spain, spain2)
        self.assertEqual(spain, spain3)
        self.assertEqual(czech.official_name, 'Czech Republic')
        self.assertEqual(bad, None)
        self.assertEqual(bad_similar, None)

    @mock.patch('scrapbag.geo.Nominatim')
    def test_get_location(self, mock_nominatim):
        """
        Test get_location
        """

        mock_nominatim.return_value.geocode.return_value = MagicMock(
            latitude=2.111, longitude=-1.222)
        coordinates = get_location('test_address')
        self.assertTrue(isinstance(coordinates, tuple))
        self.assertEqual(len(coordinates), 2)

        mock_nominatim.return_value.geocode.side_effect = Exception(
            "Error in geocode method")
        coordinates = get_location('test_address')

        self.assertEqual(coordinates, None)

    @mock.patch('scrapbag.geo.Nominatim')
    def test_get_address(self, mock_nominatim):
        """
        Test get_address
        """
        mock_nominatim.return_value.reverse.return_value = MagicMock(
            address="test address")
        address = get_address((2.111, -1.222))
        self.assertTrue(isinstance(address, str))
        self.assertEqual(address, "test address")

        # test latitude & longitude args
        address = None
        address = get_address(latitude=2.111, longitude=-1.222)
        self.assertEqual(address, "test address")

        # test location arg
        address = None
        address = get_address(location=(2.111, -1.222))
        self.assertEqual(address, "test address")

        # test only latitude
        address = None
        address = get_address(latitude=2.111)
        self.assertEqual(address, "test address")

        # test raise exception
        mock_nominatim.return_value.reverse.side_effect = Exception(
            "Error in geocode method")
        address = get_address('test_address')
        self.assertEqual(address, None)



# -*- coding: utf-8 -*-
"""
Scrapbag files file.
"""
import re
import os
import time
import zipfile
import requests

import structlog

from .collections import force_list

logger = structlog.getLogger(__name__)


def open_filezip(file_path, find_str):
    """
    Open the wrapped file.
    Read directly from the zip without extracting its content.
    """
    if zipfile.is_zipfile(file_path):
        zipf = zipfile.ZipFile(file_path)
        interesting_files = [f for f in zipf.infolist() if find_str in f]

        for inside_file in interesting_files:
            yield zipf.open(inside_file)


def extract_filezip(path_to_file, dest_path, target_zipfiles=None):
    """
    Extract file zip to destiny path folder targeting only some kind of files.
    """

    target_zipfiles = ['.*'] if target_zipfiles is None else target_zipfiles

    files = []
    _, ext = os.path.splitext(path_to_file)

    if ext == '.zip':
        file = open(path_to_file, 'rb')
        with zipfile.ZipFile(file) as zip_file:
            regexp = '|'.join(target_zipfiles) if target_zipfiles else '.*'
            search_regex = re.compile(regexp)

            lista = [m.group() for x in zip_file.namelist()
                     for m in [search_regex.search(x)] if m]

            for zp_file in lista:
                try:
                    zip_file.extract(zp_file, dest_path)
                    files.append(os.path.join(dest_path, zp_file))
                except Exception as ex:
                    msg = 'Fail to extract {} in {} to {} - {}'.format(
                        zp_file, path_to_file, dest_path, ex)
                    logger.error(msg)
        file.close()
    else:
        logger.warning('Not zipfile passed in args')
    return files


def download_file(url, destination, **kwargs):
    """
    Download file  process:
        - Open the url
        - Check if it has been downloaded and it hanged.
        - Download it to  the destination folder.

    Args:
        :urls: url to take the file.
        :destionation: place to store the downloaded file.
    """
    web_file = open_remote_url(url, **kwargs)
    file_size = 0

    if not web_file:
        logger.error(
            "Remote file not found. Attempted URLs: {}".format(url))
        return

    modified = is_remote_file_modified(web_file, destination)
    if modified:
        logger.info("Downloading: " + web_file.url)
        file_size = copy_remote_file(web_file, destination)
    else:
        logger.info("File up-to-date: " + destination)

    web_file.close()
    return file_size


def open_remote_url(urls, **kwargs):
    """Open the url and check that it stores a file.
    Args:
        :urls: Endpoint to take the file
    """
    if isinstance(urls, str):
        urls = [urls]
    for url in urls:
        try:
            web_file = requests.get(url, stream=True, **kwargs)
            if 'html' in web_file.headers['content-type']:
                raise ValueError("HTML source file retrieved.")
            return web_file
        except Exception as ex:
            logger.error('Fail to open remote url - {}'.format(ex))
            continue


def is_remote_file_modified(web_file, destination):
    """
    Check if online file has been modified.
    Args:
        :web_file: online file to check.
        :destination: path of the offline file to compare.
    """
    try:
        # check datetime of last modified in file.
        last_mod = web_file.headers.get('last-modified')
        if last_mod:
            web_file_time = time.strptime(
                web_file.headers.get(
                    'last-modified'), '%a, %d %b %Y %H:%M:%S %Z')
        else:
            web_file_time = time.gmtime()

        web_file_size = int(web_file.headers.get('content-length', -1))
        if os.path.exists(destination):
            file_time = time.gmtime(os.path.getmtime(destination))
            file_size = os.path.getsize(destination)
            if file_time >= web_file_time and file_size == web_file_size:
                return False

    except Exception as ex:
        msg = ('Fail checking if remote file is modified default returns TRUE'
               ' - {}'.format(ex))
        logger.debug(msg)

    return True


def copy_remote_file(web_file, destination):
    """
    Check if exist the destination path, and copy the online resource
    file to local.

    Args:
        :web_file: reference to online file resource to take.
        :destination: path to store the file.
    """
    size = 0
    dir_name = os.path.dirname(destination)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    with open(destination, 'wb') as file_:
        chunk_size = 8 * 1024
        for chunk in web_file.iter_content(chunk_size=chunk_size):
            if chunk:
                file_.write(chunk)
                size += len(chunk)
    return size


def remove_file(paths):
    """
    Remove file from paths introduced.
    """

    for path in force_list(paths):
        if os.path.exists(path):
            os.remove(path)



# -*- coding: utf-8 -*-
"""
Mock response module
"""


class MockRequestResponse:
    """
    Class to mock request responses, httpretty doesn't work with proxy..
    """

    def __init__(self, body, json=None, params=None, headers=None, path="",
                 status_code=200, encoding='utf-8'):

        self.status_code = status_code
        self.headers = headers

        self.text = body
        self.params = params

        self._path = path
        self._encoding = encoding

        self._json = json

    def encoding(self):
        """
        Encoding atribute getter
        Returns: encoding introduced in MockRequestResponse class
        """
        return self._encoding

    def path(self):
        """
        Path atribute getter.
        Returns: path introduced in MockRequestResponse class
        """
        return self._path

    def json(self):
        """
        Json atribute getter
        Returns: json introduced in MockRequestResponse class
        """
        return self._json



# -*- coding: utf-8 -*-
"""
Test Scrapbag collection file
"""
import unittest
from collections import OrderedDict
from scrapbag.collections import (

    exclude_values,
    exclude_empty_values,
    check_fields,
    get_element,
    add_element,
    find_value_in_object,
    force_list,
    flatten,
    merge_dicts,
    nested_dict_to_list,
    remove_list_duplicates,
    dict2orderedlist,
    get_dimension,
    get_ldict_keys,
    get_alldictkeys,
    clean_dictkeys,
    )


class UtilsCollectionsTestCase(unittest.TestCase):
    """
    Test Collection utils test case.
    """

    obj = {
        'a': 'a1',
        'aa': None,
        'b': ['b1', 'b2'],
        'c': ['c1', 'c2', 'c3'],
        'd': 'd1',
        'e': 'e1',
        'f': [{'f1': 'ff1', 'f2': 'ff2', 'g': 'ff3'}],
        'g': 'g1'
    }

    # ###### Manipulate collections functions ############################

    def test_exclude_values(self):
        """
        Test exclude_values
        """
        values = [None, [], '']
        original = {
            'a': None,
            'b': [],
            'c': '',
            'd': False,
            'e': 0,
            'f': [
                {},
                {'fa': 1, 'fb': []}],
        }

        empty_values_filter = exclude_values([], original)
        self.assertEqual(original, empty_values_filter)

        filtered = exclude_values(values, original)
        filtered2 = exclude_empty_values(original)
        self.assertEqual(len(filtered), 3)
        self.assertEqual(len(filtered), len(filtered2))
        self.assertNotEqual(filtered.get('f', {}), filtered2.get('f', {}))
        self.assertEqual(len(filtered.get('f', {})), 2)
        self.assertEqual(len(filtered2.get('f', {})), 1)

    def test_check_fields(self):
        """
        Test check_fields
        """
        fields = ('year', 'month', 'day')
        self.assertTrue(check_fields(
            fields, {'day': 1, 'year': 2014, 'month': 1}))

        self.assertFalse(check_fields(
            fields, {'day': 1, 'month': 1}))

    def test_get_element(self):
        """
        Test get_element
        """
        self.assertEqual(get_element(None, ''), None)
        self.assertEqual(get_element({'start': {'day': 1, }}, 'start/day'), 1)
        self.assertEqual(get_element({'start': {'day': 1, }}, 'start.day'), 1)
        self.assertEqual(get_element(
            {'start': {'day': 1, }}, 'start.'), {'day': 1, })
        self.assertEqual(
            get_element({'start': {'day': 1, }}, 'start.maria'), None)
        self.assertEqual(get_element(
            {'start': {'day': {'name': "Monday", 'num': 1}, }},
            'start/day.num'), 1)
        self.assertEqual(get_element({'start': [0, 1, 2]}, 'start.1'), 1)

    def test_add_element(self):
        """
        Test add_element
        """
        self.assertEqual(add_element(None, '', None), False)

        result = {}
        self.assertEqual(add_element(result, 'a', 1), {'a': 1})
        self.assertEqual(result.get('a'), 1)
        self.assertEqual(len(result.keys()), 1)

        # a is not a dict...
        self.assertEqual(add_element(result, 'a.aa', 2),  {'a': {'aa': 2}})

        # the separator keys not retrieve anything or empty path apply the
        # separator expr.
        self.assertEqual(add_element(result, '', 2, r'-'), result)

        # TEST navigate and create a key - value
        result['a'] = {}

        self.assertEqual(add_element(result, 'a.aa', 2), {'a': {'aa': 2}})
        self.assertEqual(result.get('a', {}).get('aa'), 2)
        self.assertEqual(len(result.get('a').keys()), 1)

        self.assertEqual(add_element(result, 'a.ab', 3),
                         {'a': {'aa': 2, 'ab': 3}})
        self.assertEqual(result.get('a', {}).get('aa'), 2)
        self.assertEqual(result.get('a', {}).get('ab'), 3)
        self.assertEqual(len(result.get('a').keys()), 2)

        # TEST Add to list
        result['a']['aa'] = []

        self.assertEqual(add_element(result, 'a.aa', 2),
                         {'a': {'aa': [2], 'ab': 3}})
        self.assertEqual(add_element(result, 'a.aa', 4),
                         {'a': {'aa': [2, 4], 'ab': 3}})

        self.assertEqual(add_element({'a': [1, 3]}, 'a.1', 2, override=False, digit=True), {'a': [1, 3, 2]})

        self.assertEqual(add_element({'a': [1, 3]}, 'a.1', 2, override=True, digit=True), {'a': [1, 2]})

        # fail override mode not activated and a.1 element is not a list.
        self.assertEqual(add_element(
            {'a': [1, 3]}, 'a.1', 2, override=True),
            {'a': [1, 2]})
        self.assertEqual(add_element(
            {'a': [1, 3]}, 'a.1', 2, override=False),
            {'a': [1, 3, 2]})
        self.assertEqual(add_element(
            {'a': [1, 3]}, 'a.1.0', 2),
            {'a': [1, [2]]})

        self.assertEqual(add_element(
            {'a': [1, 3]}, 'a.1.0', 2, override=True),
            {'a': [1, [2]]})

        self.assertEqual(add_element(
            {'a': [1, 3]}, 'a.1.test', 2, override=True),
            {'a': [1, {'test': 2}]})

        self.assertEqual(add_element(
            {'a': [1, 3]}, 'a.1.test.0', 2, override=True),
            {'a': [1, {'test': [2]}]})

        self.assertEqual(add_element(
            {'a': [1, 3]}, 'a.1.test.0', 2, digit=False, override=True),
            {'a': {'1': {'test': {'0': 2}}}})

        self.assertEqual(
            add_element({'a': [1, 3]}, 'a.1.test.1', 2, override=True),
            {'a': [1, {'test': ["", 2]}]})

        self.assertEqual(add_element(
            {'a': [1, {'test': [2]}]}, 'a.1.test.1', 3),
            {'a': [1, {'test': [2, 3]}]})

        self.assertEqual(
            add_element({'a': [1, {'test': [2]}]}, 'a.2.test.1', 3),
            {'a': [1, {'test': [2]}, {'test': ["", 3]}]})

        # TEST Add to dict
        result['a']['aa'] = {}

        self.assertEqual(
            add_element(result, 'a.aa', {'aaa': 2}),
            {'a': {'aa': {'aaa': 2}, 'ab': 3}})

        self.assertEqual(
            add_element(result, 'a.aa', {'aab': 3}),
            {'a': {'aa': {'aaa': 2, 'aab': 3}, 'ab': 3}})

    def test_find_value_in_nested_dict(self):
        """
        Test find_value_in_nested_dict
        """
        self.assertEqual(
            list(find_value_in_object('g', self.obj)), ['g1', 'ff3'])
        self.assertEqual(
            list(find_value_in_object('c', self.obj)), ['c1', 'c2', 'c3'])

    def test_force_list(self):
        """
        Test force_list
        """
        result = force_list(None)
        self.assertEqual(result, [])

        result = force_list([])
        self.assertEqual(result, [])

        result = force_list(['casa'])
        self.assertEqual(result, ['casa'])

        result = force_list('casa')
        self.assertEqual(result, ['casa'])

    def test_flatten(self):
        """
        Test flatten
        """
        test = OrderedDict(
            {'a': 2,
             'b': 2,
             'c': {'ca': 3, 'cb': 'test1', 'cd': [1, 3, 4], 'cc': {'cca': 1}}})

        test2 = [
            test,
            {'x': 4, 'y': 5},
            {'z': {'za': 6}}

        ]

        test3 = {'a': [
                {'aa': {'aaa': [1, 2, 3]},
                 'ab': {'aba': [1, 2, 3]},
                 'ac': {'aca': [1, 2, 3]}}
                 ]}

        test_result_1a = OrderedDict([
            ('a', 2),
            ('b', 2),
            ('c/ca', 3),
            ('c/cb', test.get('c', {}).get('cb')),
            ('c/cc/cca', 1),
            ('c/cd/1', 3),
            ('c/cd/0', 1),
            ('c/cd/2', 4)])

        test_result_1b = OrderedDict([
            ('test_a', 2),
            ('test_c_ca', 3),
            ('test_c_cd_1', 3),
            ('test_c_cd_2', 4),
            ('test_c_cd_0', 1),
            ('test_c_cb', test.get('c', {}).get('cb')),
            ('test_c_cc_cca', 1),
            ('test_b', 2)])

        test_result_2a = OrderedDict([
            ('0/a', 2),
            ('0/c/ca', 3),
            ('0/c/cb', 'test1'),
            ('0/c/cc/cca', 1),
            ('0/c/cd/0', 1),
            ('0/c/cd/2', 4),
            ('0/c/cd/1', 3),
            ('0/b', 2),
            ('2/z/za', 6),
            ('1/x', 4),
            ('1/y', 5)])

        test_result_2b = OrderedDict([
            ('test_0_a', 2),
            ('test_0_c_ca', 3),
            ('test_0_c_cb', 'test1'),
            ('test_0_c_cc_cca', 1),
            ('test_0_c_cd_0', 1),
            ('test_0_c_cd_2', 4),
            ('test_0_c_cd_1', 3),
            ('test_0_b', 2),
            ('test_2_z_za', 6),
            ('test_1_x', 4),
            ('test_1_y', 5)])

        test_result_3a = OrderedDict([
            ('a/0/ac/aca/0', 1),
            ('a/0/ac/aca/2', 3),
            ('a/0/ac/aca/1', 2),
            ('a/0/ab/aba/0', 1),
            ('a/0/ab/aba/2', 3),
            ('a/0/ab/aba/1', 2),
            ('a/0/aa/aaa/0', 1),
            ('a/0/aa/aaa/2', 3),
            ('a/0/aa/aaa/1', 2)])

        test_result_3b = OrderedDict([
            ('test_a_0_ac_aca_0', 1),
            ('test_a_0_ac_aca_2', 3),
            ('test_a_0_ac_aca_1', 2),
            ('test_a_0_ab_aba_0', 1),
            ('test_a_0_ab_aba_2', 3),
            ('test_a_0_ab_aba_1', 2),
            ('test_a_0_aa_aaa_0', 1),
            ('test_a_0_aa_aaa_2', 3),
            ('test_a_0_aa_aaa_1', 2)])

        # results of test 1
        result_1a = flatten(test, '', sep='/')
        result_1b = flatten(test, 'test')

        # results of test 2
        result_2a = flatten(test2, '', sep='/')
        result_2b = flatten(test2, 'test')

        # results of test 3
        result_3a = flatten(test3, '', sep='/')
        result_3b = flatten(test3, 'test')

        for retrieved_result, test_result in [
            (result_1a, test_result_1a),
            (result_1b, test_result_1b),
            (result_2a, test_result_2a),
            (result_2b, test_result_2b),
            (result_3a, test_result_3a),
            (result_3b, test_result_3b)
        ]:

            for key, value in test_result.items():
                self.assertIn(key, retrieved_result)
                self.assertEqual(retrieved_result[key], value)

        self.assertEqual(flatten('test'), 'test')

    def test_merge_dicts(self):

        self.assertEqual(merge_dicts({}, {'a': 1}), {'a': 1})

        self.assertEqual(merge_dicts({'a': 2}, {'a': 1}), {'a': 1})

        self.assertEqual(merge_dicts({'a': {'aa': 1}}, {'a': {'ab': 2}}), {'a': {'aa': 1, 'ab': 2}})

        self.assertEqual(merge_dicts({'a': {'aa': [3]}}, {'a': {'aa': 5}}, digit=False), {'a': {'aa': [3, 5]}})

        self.assertEqual(merge_dicts({'a': {'aa': [3]}}, {'a': {'aa': {'1': 5}}}, digit=True), {'a': {'aa': [3, 5]}})

        self.assertEqual(merge_dicts({'a': {'aa': [3]}}, {'a': {'aa': [5]}}, override=False, digit=True), {'a': {'aa': [3, 5]}})

        self.assertEqual(merge_dicts({'a': {'aa': [3]}}, {'a': {'aa': [5]}}, override=True, digit=True), {'a': {'aa': [5]}})


        # bad arguments not dicts
        self.assertEqual(merge_dicts([], {'a': 1}), {})
        self.assertEqual(merge_dicts([], []), {})
        self.assertEqual(merge_dicts({'a': 1}, []), {})

    def test_nested_dict_to_list(self):
        """
        Test nested_dict_to_list
        """
        expected = [
            ['', 'b', 'bv2'],
            ['', 'a', 'av1'],
            ['c', 'cc', 'ccv1'],
            ['c', 'ca', 'cav1'],
            ['c', 'cb', 'cbv1'],
            ['ccca', 'ccaa', 'ccaav1']]

        test = OrderedDict([
            ('b', 'bv2'),
            ('a', 'av1'),
            ('c', OrderedDict([
                ('cc', 'ccv1'),
                ('ca', 'cav1'),
                ('cb', 'cbv1'),
                ('cca', OrderedDict([
                    ('ccaa', 'ccaav1')])),
            ]))])

        result = nested_dict_to_list("", test)
        result_exclusion = nested_dict_to_list("", test, exclusion=['c'])

        self.assertEqual(
            all(row_result in expected for row_result in result), True)
        self.assertEqual(len(result), len(expected))

        check = (row_result in expected[:2] for row_result in result_exclusion)

        self.assertEqual(all(check), True)

        self.assertEqual(len(result_exclusion), len(expected[:2]))

    def test_remove_duplicates(self):
        """
        Test remove_duplicates
        """

        expected = [[1], [2, 1, 3], [2, 1, 5], [5], [88]]
        expected1 = [[2, 1, 3], [2, 1, 5], [5], [88]]
        lista = [[1], [2, 1, 3], [1], [2, 1, 5], [5], [88], [1]]

        result = remove_list_duplicates(lista)
        result1 = remove_list_duplicates(lista, unique=True)

        self.assertEqual(result, expected)
        self.assertEqual(result1, expected1)

    def test_dict2orderedlist(self):
        """
        Test dict2orderedlist
        """

        test_dict = {'a': 1, 'b': 2, 'c': 3}
        test_flatten = {'a': 1, 'b': {'ba': 2}, 'c': 3}

        result1 = dict2orderedlist(test_dict, ['a', 'b', 'c'])
        result2 = dict2orderedlist(test_dict, ['a', 'c'])
        result3 = dict2orderedlist(test_dict, ['c', 'b', 'a'])
        result4 = dict2orderedlist(
            test_flatten, ['c', 'b.ba', 'a', 'x'], default='test')

        self.assertEqual(result1, [1, 2, 3])
        self.assertEqual(result2, [1, 3])
        self.assertEqual(result3, [3, 2, 1])
        self.assertEqual(result4, [3, 2, 1, 'test'])

    def test_get_dimension(self):
        """
        Test get_dimension
        """

        test = [1, 2, 3]
        test2 = [test, test, test]
        test3 = [test2, test, test2]
        test4 = {'a': 1, 'b': 2, 'c': 3}
        bad_test = "bad_test"

        result1 = get_dimension(test)
        result2 = get_dimension(test2)
        result3 = get_dimension(test3)
        result4 = get_dimension(test4)
        result_bad_test = get_dimension(bad_test)

        self.assertEqual(result1, [3, 1])
        self.assertEqual(result2, [3, 3])
        self.assertEqual(result3, [3, 3])  # could need a fix this case..
        self.assertEqual(result4, [1, 3])
        self.assertEqual(result_bad_test, [0, 0])

    def test_get_ldict_keys(self):
        """
        Test get ldictkeys
        """
        test = [{'a': 2}, {'b': 3}, ['test'], {'c': {'ca': 4, 'cb': 5}}]
        assert_result1 = ['a', 'b', 'c']
        assert_result2a = ['a', 'b', 'c_ca', 'c_cb']
        assert_result2b = ['a', 'b', 'c.ca', 'c.cb']

        result1 = get_ldict_keys(test)
        result2a = get_ldict_keys(test, flatten_keys=True)
        result2b = get_ldict_keys(test, flatten_keys=True, sep='.')

        self.assertEqual(all(x in result1 for x in assert_result1), True)
        self.assertEqual(len(result1), 3)

        self.assertEqual(all(x in result2a for x in assert_result2a), True)
        self.assertEqual(len(result2a), 4)
        self.assertEqual(all(x in result2b for x in assert_result2b), True)
        self.assertEqual(len(result2b), 4)

    def test_get_alldictkeys(self):
        """
        Test get_alldictkeys
        """

        test = {'a': 1, 'b': 2, 'c': {'ca': 3, 'cb': 4}}
        test2 = ['a', 'b', 'c']

        assert_result1 = [('c', 'ca'), ('c', 'cb'), ('a',), ('b',)]
        result1 = get_alldictkeys(test)
        result2 = get_alldictkeys(test2)

        self.assertEqual(all(x in result1 for x in assert_result1), True)
        self.assertEqual(len(result1), 4)
        self.assertEqual(result2, [()])

    def test_clean_dictkeys(self):
        """
        Test clean_dictkeys
        """

        test = OrderedDict({
            'a': 1,
            'b': 2,
            'c': {
                'ca': 3,
                'cb': 4,
                'cd': 5},
            'd': [1, 2, {'dda': 6, 'dd': 7}]})

        result1 = clean_dictkeys([])
        assert_result1 = {}

        result2 = clean_dictkeys(test)
        assert_result2 = test

        result3 = clean_dictkeys(test, ['a'])
        assert_result3 = OrderedDict({
            'b': 2,
            'c': {
                'c': 3,
                'cb': 4,
                'cd': 5},
            'd': [1, 2, {
                'dd': [6, 7]}]})

        test = OrderedDict({
            'a': 1,
            'b': 2,
            'c': {
                'ca': 3,
                'cb': 4,
                'cd': 5},
            'd': [1, 2, {'dda': 6, 'dd': 7}]})

        result4 = clean_dictkeys(test, ['a', 'b'])
        assert_result4 = OrderedDict(
            {'c': {'c': [3, 4], 'cd': 5, }, 'd': [1, 2, {'dd': [6, 7]}]})

        self.assertEqual(result1, assert_result1)
        self.assertEqual(result2, assert_result2)
        self.assertEqual(
            sorted(flatten(dict(result3)).keys()),
            sorted(flatten(assert_result3).keys())
        )

        self.assertEqual(
            sorted(flatten(dict(result4)).keys()),
            sorted(flatten(assert_result4).keys())
        )
